;;
;; CCAT (Computational CATegory Language)
;;
;;	By Tatsuya Hagino
;;
;;	April 1984 ----> 16 May 1984 (First version)
;;
;;      Window mode	 13 June 1984, 29 November 1984
;;
;;      Improved	 September 1985,

(declare (special $stand-alone $erflag))
(eval-when (load) (setq $stand-alone t))
(cond ((not (boundp '$stand-alone)) (setq $stand-alone nil)))

(setq $erflag nil)
;for debugging
;(setq $erflag t)

(declare (macros t))

(def pp-inc (macro (l) `(setq ,(cadr l) (1+ ,(cadr l)))))

(def pp-dec (macro (l) `(setq ,(cadr l) (1- ,(cadr l)))))

(def pp-cons (macro (l) `(setq ,(caddr l) (cons ,(cadr l) ,(caddr l)))))

(def pp-cdr (macro (l) `(setq ,(cadr l) (cdr ,(cadr l)))))

(def pp-while
 (macro (l)
  `(prog () pp-while-label (cond ,(append (cdr l) '((go pp-while-label)))))))

(def pp-do
 (macro (l)
  (prog (p c vars inits)
	(pp-cdr l)
	(pp-while l
	 (cond ((or (eq (car l) 'until) (eq (car l) 'while))
		(cond ((null (cdr l)) (error 'pp-do))
		      ((eq (car l) 'until) (setq c (cadr l)))
		      (t (setq c `(not ,(cadr l)))))
		(pp-cons `(cond (,c (return))) p)
		(pp-cdr l))
	       ((eq (car l) 'var)
		(cond ((or (null (cdr l)) (not (atom (cadr l))))
		       (error 'pp-do)))
		(pp-cons (cadr l) vars))
	       ((eq (car l) 'vars)
		(cond ((or (null (cdr l)) (not (dtpr (cadr l))))
		       (error 'pp-do)))
		(setq vars (append (cadr l) vars)))
	       ((eq (car l) 'init)
		(cond ((null (cdr l)) (error 'pp-do)))
		(pp-cons (cadr l) inits))
	       (t (pp-cons (car l) p)))
	 (pp-cdr l))
	(return (cons 'prog
		      (cons (nreverse vars)
			    (nconc (nreverse inits)
				   (cons 'pp-do-label
					 (nconc (nreverse p)
						'((go pp-do-label)))))))))))


(declare
 (special user-top-level
	  $system-mode
	  $arrow-dir
	  $w-show
	  $w-diagram
	  $w-m-list
	  $w-o-list
	  $w-command
	  $w-edit
	  $w-ll
	  $w-help
	  $w-help-on
	  $simp-mode
	  $aliases
	  $keypad))

(setq $w-show nil)
(setq $w-diagram nil)
(setq $w-m-list nil)
(setq $w-o-list nil)
(setq $w-edit nil)
(setq $w-command nil)
(setq $w-help nil)
(setq $w-help-on nil)
(setq $aliases nil)
(setq $keypad nil)
(setq $arrow-dir 'normal)
(setq $simp-mode 'lazy)

;; $system-mode is either 'line or 'window
(setq $system-mode 'line)

(def nterpri
 (macro (l) '(progn (tyo 13) (terpri))))

;; lexical analyzer 

; reading routines

(declare (special $rprmpt $rtyi $rnl $rtok $ln $reof))

(setq $rprmpt "| ")
(setq $rtyi nil)
(setq $rnl nil)
(setq $rtok nil)
(setq $reof nil)
(setq $ln 70)

(def rascii
 (macro (l) (car (exploden (cadr l)))))

(def rarrow (macro (l) ''|->|))

(def rcolon (macro (l) ''|:|))

(def rcomp (macro (l) ''|.|))

(def rrcomp (macro (l) ''|\||))

(def rlpar (macro (l) ''|(|))

(def rrpar (macro (l) ''|)|))

(def rcomma (macro (l) ''|,|))

(def requal (macro (l) ''|=|))

(def rstar (macro (l) ''|*|))

(def rtyi
 (lambda (pt)
  (prog (x)
   	(cond ($rtyi (setq x (car $rtyi))
		     (setq $rtyi (cdr $rtyi))
		     (return x))
	      ($reof (return -1)))
	(cond ((and $rnl (null pt) (status isatty)) (princ $rprmpt)))
	(cond ((null pt) (setq x (tyi)))
	      ((= pt 'line) (setq x -1))
	      ((= pt 'window)
	       (cond ((null $w-ll) (setq x -1))
		     (t (setq $rtyi (car $w-ll))
			(setq $w-ll (cdr $w-ll))
			(setq x 10))))
	      (t (setq x (tyi pt))))
	(setq $reof (< x 0))
	(setq $rnl (= x 10))
	(return x))))

(def rbtyi
 (macro (x)
  `(setq $rtyi (cons ,(cadr x) $rtyi))))

(def rdigit
 (lambda (ch) (and (<= (rascii \0) ch) (<= ch (rascii \9)))))

(def rletter
 (lambda (ch) (or (and (<= (rascii \A) ch) (<= ch (rascii \Z)))
		  (and (<= (rascii \a) ch) (<= ch (rascii \z)))
		  (= ch (rascii \!)))))

(def rspecial
 (lambda (ch) (or (= ch (rascii \,))
		  (= ch (rascii \.))
		  (= ch (rascii \)))
		  (= ch (rascii \())
		  (= ch (rascii \:))
		  (= ch (rascii \;)))))

(def rtok
 (lambda (pt)
  (prog (ch chl)
	(cond ($rtok (setq ch (car $rtok))
		     (setq $rtok (cdr $rtok))
		     (return ch)))
   aa   (setq ch (rtyi pt))
	(cond ((< ch 0) (return -1))
	      ((<= ch (rascii \ )) (go aa))
	      ((rspecial ch) (return (ascii ch)))
	      ((rdigit ch) (go cc))
	      ((rletter ch) (go cc)))
   dd   (setq chl (cons ch chl))
	(setq ch (rtyi pt))
	(cond ((and (not (<= ch (rascii \ )))
		    (not (rspecial ch))
		    (not (rdigit ch))
		    (not (rletter ch)))
	       (go dd)))
   ee   (rbtyi ch)
	(return (intern (maknam (reverse chl))))
   cc   (setq chl (cons ch chl))
	(setq ch (rtyi pt))
	(cond ((or (rletter ch) (rdigit ch)
		   (= ch (rascii \_))) (go cc)))
	(go ee))))

(def rtokl
 (lambda (pt)
  (prog (tok tokl)
   aa   (setq tok (rtok pt))
	(cond ((or (eq tok '\;) (eq tok -1)) (return (reverse tokl))))
	(setq tokl (cons tok tokl))
	(go aa))))

(def rclear
 (lambda ()
  (progn (setq $rtyi nil)
	 (setq $reof nil)
	 (setq $rtok nil))))

;; paser

(def oread
 (lambda (tokl)
  (prog (fn pt ok)
	(cond ((null tokl) (princ "No file name") (nterpri) (return nil)))
	(setq fn (concatl tokl))
	(cond ((null (probef fn))
	       (princ "File dose not exist") (nterpri) (return nil)))
	(rclear)
	(setq pt (fileopen fn "r"))
   aa   (setq tokl (rtokl pt))
	(cond ((null tokl) (setq ok nil))
	      ((or (eq (car tokl) 'right) (eq (car tokl) 'left))
	       (setq ok (object tokl)))
	      ((eq (car tokl) 'let)
	       (setq ok (maux (cdr tokl))))
	      (t (princ "Unexpected input: ")
		 (princ (car tokl))
		 (nterpri)
		 (setq ok nil)))
	(cond (ok (go aa)))
	(close pt)
	(rclear)
	(return nil))))

; paser for objects

(declare (special $error $errpos $object $morphism))

(setq $object nil)
(setq $morphism nil)

(def oerrposprint
 (lambda (tokl etokl)
  (prog (oktokl cnt)
	(cond ((null tokl) (return nil)))
   aa   (cond ((null tokl) (go bb))
	      ((equal tokl etokl) (go bb)))
	(setq oktokl (cons (car tokl) oktokl))
	(setq tokl (cdr tokl))
	(go aa)
   bb   (setq cnt 0)
	(setq tokl nil)
   cc   (cond ((and oktokl (< cnt 5))
	       (setq tokl (cons (car oktokl) tokl))
	       (setq oktokl (cdr oktokl))
	       (setq cnt (1+ cnt))
	       (go cc)))
	(princ "... ")
   dd   (cond (tokl (princ (car tokl))
		    (princ " ")
		    (setq tokl (cdr tokl))
		    (go dd)))
	(princ "??? ")
	(setq cnt 0)
   ee   (cond ((and etokl (< cnt 5))
	       (princ (car etokl))
	       (princ " ")
	       (setq etokl (cdr etokl))
	       (setq cnt (1+ cnt))
	       (go ee)))
	(princ "...")
	(nterpri))))

(def object
 (lambda (tokl)
  (prog (obj mdl nobj)
	(setq obj (errset (ohead tokl) $erflag))
	(cond ((atom obj) (princ "Syntax error: ")
			  (oprinl $error)
			  (nterpri)
			  (oerrposprint tokl $errpos)
			  (return nil)))
	(setq obj (car obj))
	(cond ((null (setq nobj (odecl (car obj) (cdr obj) nil)))
	       (return nil)))
	(odecl (setq obj (car obj)) nobj t)
	(setq mdl (ofunc obj nobj))
	(cond ((car (cddddr nobj)) (princ (car (cddddr nobj))) (princ " ")))
	(princ (car nobj))
	(princ " object ")
	(princ obj)
	(cond ((null mdl) (go bb)))
	(princ (rlpar))
   aa	(cond ((eq (car mdl) 'covariant) (princ "+"))
	      ((eq (car mdl) 'contravariant) (princ "-"))
	      ((eq (car mdl) 'free) (princ "0"))
	      (t (princ "?")))
	(setq mdl (cdr mdl))
	(cond (mdl (princ (rcomma)) (go aa)))
	(princ (rrpar))
   bb	(princ " defined")
	(nterpri)
	(return t))))

(def odecl
 (lambda (obj l flg)
  (prog (x y md1 md2 rl dom mdom productive codom para npara)
	(setq rl (car l))
	(cond (flg
	       (cond ((get obj 'object) (odelo obj nil)))
	       (setq $object (cons obj $object))
	       (putprop obj l 'object)
	       (cond ((oinco obj) (putprop obj t 'incorrect))))
	      ((and (setq y (get obj 'morphism))
		    (not (eq (cadr y) obj)))
	       (oprinl `("object" ,obj "is already declared as a morphism"))
	       (go zz)))
	(setq x (caddr l))
	(cond (flg (putprop x `(factorizer ,obj) 'morphism)
		   (setq $morphism (cons x $morphism)))
	      ((and (setq y (get x 'morphism)) (not (eq (cadr y) obj)))
	       (oprinl `("factorizer" ,x "is used in" ,(cadr y)))
	       (go zz)))
	(setq x (cadddr l))
	(cond ((eq rl 'right) (setq productive 'productive)
			      (setq mdom '("domain" "productive")))
	       (t (setq productive 'coproductive)
		  (setq mdom '("codomain" "coproductive"))))
	(setq para (cadr l))
   aa   (cond ((null x)
	       (cond (flg (return nil))
		     (t (return (append l (list productive))))))
	      (flg (putprop (caar x)
			    `(unit ,obj ,(cadar x) ,(caddar x))
			    'morphism)
		   (setq $morphism (cons (caar x) $morphism))
		   (setq x (cdr x))
		   (go aa))
	      ((and (setq y (get (caar x) 'morphism))
		    (not (eq (cadr y) obj)))
	       (oprinl `("unit" ,(caar x) "is used in" ,(cadr y)))
	       (go zz))
	      ((eq rl 'right) (setq dom (cadar x)) (setq codom (caddar x)))
	      (t (setq dom (caddar x)) (setq codom (cadar x))))
	(cond ((and productive (atom codom) (memq codom para))
	       (cond ((memq codom npara) (setq productive nil))
		     (t (pp-cons codom npara))))
	      (t (setq productive nil)))
	(cond ((not (equal (ooccur dom obj) 1))
	       (oprinl `(,obj
			 "should occur once in"
			 ,(car mdom)
			 "of"
			 ,(caar x)))
	       (go zz))
	      ((not (oprodchk dom obj rl))
	       (oprinl `(,obj
			 "should occur inside"
			 ,(cadr mdom)
			 "objects"
			 "in"
			 ,(car mdom)
			 "of"
			 ,(caar x)))
	       (go zz)))
	(setq md1 (ofunmod (cadar x) obj))
	(setq md2 (ofunmod (caddar x) obj))
	(cond ((or (eq md1 'fix) (eq md2 'fix))
	       (oprinl `(,obj "occurs inconsistently in" ,(caar x)))
	       (go zz))
	      ((or (eq md1 'contravariant) (eq md2 'contravariant))
	       (oprinl `(,obj "must not occur contravariantly in " ,(caar x)))
	       (go zz))
	      ((and (eq md1 'free) (eq md2 'free))
	       (oprinl `(,(caar x) "does not depend on" ,obj))
	       (go zz)))
	(setq x (cdr x))
	(go aa)
   zz	(nterpri)
	(return nil))))

(def ooccur
 (lambda (dom obj)
  (prog (l sum)
	(cond ((eq dom obj) (return 1))
	      ((atom dom) (return 0)))
	(setq l (caddr dom))
	(setq sum 0)
   aa   (cond ((null l) (return sum)))
	(setq sum (plus sum (ooccur (car l) obj)))
	(setq l (cdr l))
	(go aa))))

(def oprodchk
 (lambda (dom obj rl)
  (prog (l)
	(cond ((eq dom obj) (return t))
	      ((not (equal (ooccur dom obj) 1)) (return nil)))
	(setq l (get (cadr dom) 'object))
	(cond ((null l) (return nil))
	      ((not (eq (car l) rl)) (return nil)))
	(cond ((or (and (eq rl 'right) (not (eq (car (cddddr l)) 'productive)))
		   (and (eq rl 'left)
			(not (eq (car (cddddr l)) 'coproductive))))
	       (return nil)))
	(setq l (caddr dom))
	(pp-while (and l (not (equal (ooccur (car l) obj) 1))) (pp-cdr l))
	(cond ((null l) (return nil)))
	(return (oprodchk (car l) obj rl)))))

(def oprinl
 (lambda (msgl)
  (cond ((null msgl) nil)
	((atom msgl) (princ msgl))
	(t (princ (car msgl))
	   (cond ((cdr msgl) (princ " ")))
	   (oprinl (cdr msgl))))))

(def oerror
 (lambda (emsg tokl)
  (progn (setq $error emsg) (setq $errpos tokl) (error 0))))

(def oeq
 (lambda (tokl tok)
  (cond ((null tokl) (oerror `("expecting" ,tok)  tokl))
	(t (eq (car tokl) tok)))))

(def ocheck
 (lambda (tokl tok)
  (cond ((or (null tokl) (not (eq (car tokl) tok)))
	 (oerror `("expecting" ,tok) tokl)))))

(def onext
 (macro (l)
  (cond ((cdr l) `(setq ,(cadr l) (cdr ,(cadr l))))
	(t '(setq tokl (cdr tokl))))))

(def onmchk
 (lambda (tokl msg)
  (cond ((null tokl) (oerror `("no" ,msg "name") nil))
	((memq (car tokl) '(with end is it
 	                   \) \( \: \-\> \, \.))
	 (oerror `("illegal" ,msg "name") tokl))
	(t (car tokl)))))

(def ohead
 (lambda (tokl)
  (prog (name rl argl univ unit unitl)
	(cond ((oeq tokl 'right) (setq rl 'right))
	      ((oeq tokl 'left) (setq rl 'left))
	      (t (oerror "right or left expected" tokl)))
	(onext)
	(ocheck tokl 'object)
	(onext)
	(setq name (onmchk tokl "object"))
	(onext)
	(cond ((not (oeq tokl (rlpar))) (go bb)))
   aa   (onext)
	(setq argl (cons (onmchk tokl "parameter") argl))
	(cond ((memq (car argl) (cdr argl))
	       (oerror "duplicate parameters" tokl)))
	(onext)
	(cond ((oeq tokl (rcomma)) (go aa)))
	(ocheck tokl (rrpar))
	(onext)
	(setq argl (reverse argl))
	(cond ((memq name argl) (oerror "name confusion in parameters" tokl)))
   bb   (ocheck tokl 'with)
	(onext)
	(setq univ (onmchk tokl "factorizer"))
	(onext)
	(cond ((oeq tokl 'is) (onext)))
   cc   (cond ((oeq tokl 'end) (go dd)))
	(setq unit (ounit tokl name argl))
	(setq tokl (cadr unit))
	(cond ((assoc (caar unit) unitl)
	       (oerror "duplicate unit name" tokl))
	      ((eq univ (caar unit))
	       (oerror "unit name equal to factorizer" tokl)))
	(setq unitl (cons (car unit) unitl))
	(go cc)
   dd   (onext)
	(setq unitl (reverse unitl))
	(ocheck tokl 'object)
	(onext)
	(cond (tokl (oerror "somthing after end" tokl)))
	(return (list name rl argl univ unitl)))))

(def ounit
 (lambda (tokl obj argl)
  (prog (name domain codomain)
	(setq name (onmchk tokl "unit"))
	(onext)
	(ocheck tokl (rcolon))
	(onext)
	(setq domain (odom tokl (cons obj argl) nil nil))
	(setq tokl (cadr domain))
	(setq domain (car domain))
	(ocheck tokl (rarrow))
	(onext)
	(setq codomain (odom tokl (cons obj argl) nil nil))
	(setq tokl (cadr codomain))
	(setq codomain (car codomain))
	(return (list (list name domain codomain) tokl)))))

(def odom
 (lambda (tokl argl tvar tvl)
  (prog (name para fpara apara oobj)
	(cond ((and tvar (eq (car tokl) (rstar)))
	       (onext)
	       (setq name (onmchk tokl "parameter"))
	       (onext)
	       (setq name (intern (concat (rstar) name)))
	       (setq para (assq name tvl))
	       (cond ((null para)
		      (setq para (car (mnewvar '(*a))))
		      (return (list para tokl (cons (list name para) tvl))))
		     (t (return (list (cadr para) tokl tvl))))))
	(setq name (onmchk tokl "object"))
	(onext)
	(cond ((memq name argl)
	       (cond (tvar (return (list name tokl tvl)))
		     (t (return (list name tokl))))))
	(setq oobj (get name 'object))
	(cond ((null oobj) (oerror `("undefined object" ,name) tokl)))
	(setq fpara (cadr oobj))
	(cond ((null fpara) (go bb)))
	(ocheck tokl (rlpar))
   aa	(onext)
	(setq para (odom tokl argl tvar tvl))
	(setq tokl (cadr para))
	(cond (tvar (setq tvl (caddr para))))
	(pp-cons (car para) apara)
	(pp-cdr fpara)
	(cond (fpara (ocheck tokl (rcomma)) (go aa)))
	(ocheck tokl (rrpar))
	(onext)
	(setq apara (nreverse apara))
   bb   (cond (tvar (return (list (list 'object name apara) tokl tvl)))
	      (t (return (list (list 'object name apara) tokl)))))))

; functor associated with objects

(def ofunc
 (lambda (objnm obj)
  (prog (rl ul mdl argl argmdl x)
	(setq rl (car obj))
	(setq ul (cadddr obj))
   aa	(cond ((null ul) (go bb))
	      ((eq rl 'right)
	       (setq mdl (ofunmoda (cadar ul) 'contravariant mdl))
	       (setq mdl (ofunmoda (caddar ul) 'covariant mdl)))
	      (t (setq mdl (ofunmoda (cadar ul) 'covariant mdl))
		 (setq mdl (ofunmoda (caddar ul) 'contravariant mdl))))
	(setq ul (cdr ul))
	(go aa)
   bb	(setq argl (cadr obj))
   cc   (cond ((null argl) (go dd)))
	(setq x (assoc (car argl) mdl))
	(cond ((null x) (setq x 'free)) (t (setq x (cdr x))))
	(setq argmdl (cons x argmdl))
	(setq argl (cdr argl))
	(go cc)
   dd	(setq argmdl (reverse argmdl))
	(putprop objnm (list 'functor objnm nil argmdl) 'morphism)
	(return argmdl))))

(def ofunmod
 (lambda (o para)
  (prog (ml x)
	(setq ml (ofunmoda o 'covariant nil))
	(setq x (assoc para ml))
	(cond ((null x) (return 'free))
	      (t (return (cdr x)))))))

(def ofunmoda
 (lambda (o cm ml)
  (cond ((atom o) (ofunmodaa o cm ml))
	(t (prog (ml1 argl)
		 (setq ml1 (cadddr (get (cadr o) 'morphism)))
		 (setq argl (caddr o))
	    aa   (cond ((null ml1) (return ml))
		       ((or (eq cm 'fix) (eq (car ml1) 'fix))
		        (setq ml (ofunmoda (car argl) 'fix ml)))
		       ((or (eq cm 'free) (eq (car ml1) 'free)) nil)
		       ((eq cm 'covariant)
		        (setq ml (ofunmoda (car argl) (car ml1) ml)))
		       ((eq (car ml1) 'covariant)
			(setq ml (ofunmoda (car argl) 'contravariant ml)))
		       (t (setq ml (ofunmoda (car argl) 'covariant ml))))
		 (setq ml1 (cdr ml1))
		 (setq argl (cdr argl))
		 (go aa))))))

(def ofunmodaa
 (lambda (o cm ml)
  (cond ((null ml) (list (cons o cm)))
	((eq (caar ml) o)
	 (cond ((eq cm 'free) ml)
	       ((eq (cdar ml) 'free) (cons (cons o cm) (cdr ml)))
	       ((eq cm (cdar ml)) ml)
	       (t (cons (cons o 'fix) (cdr ml)))))
	(t (cons (car ml) (ofunmodaa o cm (cdr ml)))))))

; paser for morphisms

(declare (special $var $avar))

(putprop 'id '(identity *identity *a *a) 'morphism)

(def maux
 (lambda (tokl)
  (prog (m y)
	(setq m (errset (mauxp tokl) $erflag))
	(cond ((atom m) (princ "Syntax error: ")
			(oprinl $error)
			(nterpri)
			(oerrposprint tokl $errpos)
			(return nil)))
	(setq m (car m))
	(cond ((and (setq y (get (car m) 'morphism))
		    (not (eq (car y) 'auxiliary)))
	       (oprinl `("morphism" ,(car m) "is already defined as" ,(car y)))
	       (nterpri)
	       (return nil))
	      (y (odelm (car m) nil)))
	(putprop (car m)
		 (cons 'auxiliary
		       (cons '*auxiliary (cdr m)))
		 'morphism)
	(cond ((oincml (cadr m)) (putprop (car m) t 'incorrect))
	      (t (remprop (car m) 'incorrect)))
	(cond ((not (memq (car m) $morphism))
	       (setq $morphism (cons (car m) $morphism))))
	(pmobj (car m) (cadddr m) (car (cddddr m)))
	(princ " defined")
	(nterpri)
	(return t))))

(def mauxp
 (lambda (tokl)
  (prog (mn m dom codom vl avl)
	(setq mn (onmchk tokl "morphism"))
	(onext)
	(or (oeq tokl 'be) (ocheck tokl (requal)))
	(onext)
	(setq m (morph0 tokl nil))
	(return (cons mn m)))))

(def morph0
 (lambda (tokl expand)
  (prog (m vl avl dom codom)
	(setq $var 0)
	(setq m (morph tokl nil expand))
	(cond ((cadddr m) (oerror `(,(rcomp) "expected") tokl)))
	(cond ((not (equal $var 0))
	       (setq vl (mgetvo (cadr m) (mgetvo (caddr m) nil)))
	       (setq $avar nil)
	       (setq avl (mnewvara vl nil))
	       (setq vl (mpair vl avl))))
	(cond ((null vl) (setq dom (cadr m))
			 (setq codom (caddr m))
			 (setq m (car m)))
	      (t (setq dom (msubsto (cadr m) vl))
		 (setq codom (msubsto (caddr m) vl))
		 (setq m (car m))))
	(return (list m avl dom codom)))))

(def morph
 (lambda (tokl tvl expand)
  (prog (obj x y vl nm dc m ml dom codom)
	(cond ((null tokl) (oerror "illegal end" nil)))
	(setq nm (car tokl))
	(onext)
	(setq dc (get nm 'morphism))
	(cond ((null dc) (oerror `("no morphism" ,nm) tokl))
	      ((eq (car dc) 'identity)
	       (cond ((or (null tokl) (not (eq (car tokl) (rlpar))))
		      (setq dom (car (mnewvar '(*a)))))
		     (t (onext)
			(setq dom (odom tokl nil t tvl))
			(setq tvl (caddr dom))
			(setq tokl (cadr dom))
			(setq dom (car dom))
			(ocheck tokl (rrpar))
			(onext)))
	       (setq m (list (list 'identity)))
	       (setq codom dom))
	      ((eq (car dc) 'unit)
	       (setq obj (get (cadr dc) 'object))
	       (setq x (cadr obj))
	       (setq vl (mnewvar x))
	       (setq y (list 'object (cadr dc) vl))
	       (setq vl (cons (list (cadr dc) y) (mpair x vl)))
	       (setq m (list (list 'unit nm)))
	       (setq x (cdr (assoc nm (cadddr obj))))
	       (setq dom (msubsto (car x) vl))
	       (setq codom (msubsto (cadr x) vl)))
	      ((eq (car dc) 'factorizer)
	       (setq dc (muniv nm (cadr dc) tokl tvl expand))
	       (setq m (list (car dc)))
	       (setq dom (cadr dc))
	       (setq codom (caddr dc))
	       (setq tokl (cadddr dc))
	       (setq tvl (car (cddddr dc))))
	      ((eq (car dc) 'auxiliary)
	       (setq vl (mnewvar (cadddr dc)))
	       (cond (expand (setq m (caddr dc)))
		     (t (setq m (list (list 'auxiliary nm)))))
	       (cond ((null vl) (setq dom (car (cddddr dc)))
				(setq codom (cadr (cddddr dc))))
		     (t (setq vl (mpair (cadddr dc) vl))
	      		(setq dom (msubsto (car (cddddr dc)) vl))
	       		(setq codom (msubsto (cadr (cddddr dc)) vl)))))
	      ((eq (car dc) 'functor)
	       (setq dc (mfun nm (cadddr dc) tokl tvl expand))
	       (setq m (car dc))
	       (setq dom (cadr dc))
	       (setq codom (caddr dc))
	       (setq tokl (cadddr dc))
	       (setq tvl (car (cddddr dc))))
	      (t (oerror "strange morphism" tokl)))
	(cond ((or (null tokl)
		   (and (not (eq (car tokl) (rcomp)))
			(not (eq (car tokl) (rrcomp)))))
	       (return (list m dom codom tokl tvl)))
	      ((eq (car tokl) (rrcomp))
	       (onext)
	       (setq ml (morph tokl tvl expand))
	       (setq vl (munify codom (cadr ml)))
	       (setq m (mcomp (car ml) m))
	       (setq dom (msubsto dom vl))
	       (setq codom (msubsto (caddr ml) vl)))
	      (t (onext)
		 (setq ml (morph tokl tvl expand))
		 (setq vl (munify dom (caddr ml)))
		 (setq m (mcomp m (car ml)))
		 (setq dom (msubsto (cadr ml) vl))
		 (setq codom (msubsto codom vl))))
	(return (list m dom codom (cadddr ml) (car (cddddr ml)))))))

(def muniv
 (lambda (univ objnm tokl tvl expand)
  (prog (m rl obj para argl unitl dom codom dom1 codom1 vl nv o oo)
	(setq obj (get objnm 'object))
	(setq rl (car obj))
	(setq para (cons objnm (cadr obj)))
	(setq unitl (mapcar 'cdr (cadddr obj)))
	(setq nv (mnewvar para))
	(setq vl (mpair para nv))
	(setq unitl (msubstoll unitl vl))
	(setq o (list 'object objnm (cdr nv)))
	(setq oo (car nv))
	(cond ((eq rl 'right) (setq dom oo) (setq codom o))
	      (t (setq dom o) (setq codom oo)))
	(cond ((null unitl)
	       (return (list (list 'factorizer univ argl)
			     dom
			     codom
			     tokl
			     tvl))))
	(ocheck tokl (rlpar))
   aa   (onext)
	(setq m (morph tokl tvl expand))
	(setq tvl (car (cddddr m)))
	(setq tokl (cadddr m))
	(setq dom1 (cadr m))
	(setq codom1 (caddr m))
	(setq m (car m))
	(setq vl (munify dom1 (caar unitl)))
	(setq codom1 (msubsto codom1 vl))
	(setq unitl (msubstoll unitl vl))
	(setq dom (msubsto dom vl))
	(setq codom (msubsto codom vl))
	(setq o (msubsto o vl))
	(setq oo (msubsto oo vl))
	(setq vl (munify codom1 (cadar unitl)))
	(setq argl (cons m argl))
	(setq unitl (msubstoll (cdr unitl) vl))
	(setq dom (msubsto dom vl))
	(setq codom (msubsto codom vl))
	(setq o (msubsto o vl))
	(setq oo (msubsto oo vl))
	(cond (unitl (ocheck tokl (rcomma)) (go aa)))
	(ocheck tokl (rrpar))
	(onext)
	(return (list (list 'factorizer univ (reverse argl))
		      dom
		      codom
		      tokl
		      tvl)))))

(def mfun
 (lambda (mn mdl tokl tvl expand)
  (prog (para para1 ml m dom codom)
	(cond ((null mdl)
	       (setq dom (list 'object mn nil))
	       (return (list (list (list 'identity)) dom dom tokl tvl))))
	(ocheck tokl (rlpar))
	(setq para (setq para1 (cadr (get mn 'object))))
   aa   (onext)
	(setq m (morph tokl tvl expand))
	(setq dom (cadr m))
	(setq codom (caddr m))
	(setq tokl (cadddr m))
	(setq tvl (car (cddddr m)))
	(setq m (car m))
	(setq ml (cons (list (car para) m dom codom) ml))
	(cond ((and (eq (car mdl) 'fix) (not (equal m `((identity)))))
	       (oerror `("argument error for" ,mn) tokl)))
	(setq mdl (cdr mdl))
	(setq para (cdr para))
	(cond (mdl (ocheck tokl (rcomma)) (go aa)))
	(ocheck tokl (rrpar))
	(onext)
	(setq ml (mfungt (list 'object mn para1) ml))
	(return (list (car ml) (cadr ml) (caddr ml) tokl tvl)))))

(def mfungt
 (lambda (o argl)
  (prog (objnm obj fargl nargl unitl rl para apara bpara f mdl unit mll oa ob
	       dom codom)
	(cond ((atom o) (return (cdr (assoc o argl)))))
	(setq objnm (cadr o))
	(setq obj (get objnm 'object))
	(setq mdl (cadddr (get objnm 'morphism)))
	(setq para (cadr obj))
	(setq fargl (caddr o))
	(setq rl (car obj))
   aa   (cond ((null para) (go bb)))
	(setq f (mfungt (car fargl) argl))
	(setq nargl (cons (cons (car para) f) nargl))
	(cond ((eq (car mdl) 'contravariant)
	       (setq apara (cons (caddr f) apara))
	       (setq bpara (cons (cadr f) bpara)))
	      (t (setq apara (cons (cadr f) apara))
		 (setq bpara (cons (caddr f) bpara))))
	(setq para (cdr para))
	(setq fargl (cdr fargl))
	(go aa)
   bb	(setq unitl (cadddr obj))
	(setq apara (list 'object objnm (reverse apara)))
	(setq bpara (list 'object objnm (reverse bpara)))
	(cond ((eq rl 'right) (setq oa apara) (setq ob bpara))
	      (t (setq oa bpara) (setq ob apara)))
	(setq nargl (cons (list objnm (list (list 'identity)) oa oa)
			   nargl))
   cc   (cond ((null unitl) (go dd)))
	(setq dom (mfungt (cadar unitl) nargl))
	(setq codom (mfungt (caddar unitl) nargl))
	(setq unit (list 'unit (caar unitl)))
	(setq mll (cons (mcomp (car codom) (mcomp (list unit) (car dom)))
		         mll))
	(setq unitl (cdr unitl))
	(go cc)
   dd   (return (list (list (list 'factorizer (caddr obj) (reverse mll)))
		      apara
		      bpara)))))

(def mnewvar
 (lambda (ovar)
  (cond ((null ovar) nil)
	(t (cons (list '*var (setq $var (add1 $var)))
	         (mnewvar (cdr ovar)))))))

(def mpair
 (lambda (x y)
  (cond ((null x) nil)
	(t (cons (list (car x) (car y)) (mpair (cdr x) (cdr y)))))))

(def msubsto
 (lambda (o vl)
  (prog (x)
	(setq x (assoc o vl))
	(cond (x (return (cadr x)))
	      ((not (eq (car o) 'object)) (return o)))
	(return (list 'object (cadr o) (msubstol (caddr o) vl))))))

(def msubstol
 (lambda (ol vl)
  (cond ((null ol) nil)
	(t (cons (msubsto (car ol) vl) (msubstol (cdr ol) vl))))))

(def msubstoll
 (lambda (oll vl)
  (cond ((null oll) nil)
	(t (cons (msubstol (car oll) vl) (msubstoll (cdr oll) vl))))))

(def msubstvl
 (lambda (vl svl)
  (cond ((null vl) nil)
	(t (cons (list (caar vl) (msubsto (cadar vl) svl))
		 (msubstvl (cdr vl) svl))))))

(def munify
 (lambda (o1 o2)
  (cond ((atom o1) (eq o1 o2))
	((eq (car o1) '*var) (list (list o1 o2)))
	((eq (car o2) '*var) (list (list o2 o1)))
	((not (eq (cadr o1) (cadr o2))) (oerror "domains incompatible" nil))
	(t (prog (vl vl1)
		 (setq o1 (caddr o1))
		 (setq o2 (caddr o2))
	    aa   (cond ((null o1) (return vl)))
		 (setq vl1 (munify (car o1) (car o2)))
		 (setq o1 (msubstol (cdr o1) vl1))
		 (setq o2 (msubstol (cdr o2) vl1))
		 (setq vl (msubstvl vl vl1))
		 (setq vl (append vl1 vl))
		 (go aa))))))

(def mgetvo
 (lambda (o vl)
  (cond ((atom o) vl)
	((eq (car o) '*var)
	 (cond ((member o vl) vl) (t (cons o vl))))
	(t (mgetvol (caddr o) vl)))))

(def mgetvol
 (lambda (ol vl)
  (cond ((null ol) vl)
	(t (mgetvo (car ol) (mgetvol (cdr ol) vl))))))

(def mnewvara
 (lambda (vl old)
  (prog (rv alf dec l nv)
	(cond ($avar (setq alf (car $avar)) (setq dec (cadr $avar)))
	      (t (setq alf (rascii a)) (setq dec (rascii 0))))
	(pp-while vl
		  (cond ((equal dec (rascii 0)) (setq l (list alf)))
			(t (setq l (list alf dec))))
		  (setq nv (implode (cons (rascii *) l)))
		  (cond ((not (memq nv old)) (pp-cons nv rv) (pp-cdr vl)))
		  (pp-inc alf)
		  (cond ((> alf (rascii z))
			 (setq alf (rascii a))
			 (pp-inc dec))))
	(setq $avar (list alf dec))
	(return rv))))

(def mcomp
 (lambda (ml1 ml2)
  (cond ((null ml1) ml2)
	(t (mcomp1 (car ml1) (mcomp (cdr ml1) ml2))))))

(def mcomp1
 (lambda (m ml)
  (cond ((null ml) (list m))
	((eq (car m) 'identity) ml)
	((eq (caar ml) 'identity) (mcomp1 m (cdr ml)))
	(t (cons m ml)))))

;; print routines

(def pmobj
 (lambda (mn dom codom)
  (progn (princ mn)
	 (princ " : ")
	 (pobj dom)
	 (princ " -> ")
	 (pobj codom))))

(def pobj
 (lambda (o)
  (prog ()
	(cond ((atom o) (princ o) (return nil)))
	(princ (cadr o))
	(setq o (caddr o))
	(cond ((null o) (return nil)))
	(princ (rlpar))
   aa   (pobj (car o))
	(setq o (cdr o))
	(cond (o (princ (rcomma)) (go aa)))
	(princ (rrpar))
	(return nil))))

(def pmorph
 (lambda (m lvl)
  (progn
   (cond ((> (nwritn) $ln) (nterpri) (princ "    ")))
   (cond ((zerop lvl) (princ `|&|))
	 ((eq (car m) 'identity) (princ "id"))
	 ((or (eq (car m) 'unit) (eq (car m) 'auxiliary)) (princ (cadr m)))
	 ((eq (car m) 'factorizer)
	  (prog (mll)
	        (princ (cadr m))
	        (setq mll (caddr m))
	        (cond ((null mll) (return nil)))
	        (princ (rlpar))
	   aa   (pmorphl (car mll) (sub1 lvl))
	        (setq mll (cdr mll))
	        (cond ((null mll) (princ (rrpar)) (return nil)))
	        (princ (rcomma))
	        (go aa)))
	 (t (oerror "strange morphism" nil))))))

(def pmorphl
 (lambda (ml lvl)
  (cond ((null ml) (oerror "strange morphism" nil))
	((eq $arrow-dir 'reverse)
	 (cond ((cdr ml) (pmorphl (cdr ml) lvl) (princ (rrcomp))))
	 (pmorph (car ml) lvl))
	(t (pmorph (car ml) lvl)
	   (cond ((cdr ml) (princ (rcomp)) (pmorphl (cdr ml) lvl)))))))

(def oshow
 (lambda (tokl)
  (prog (ml dom codom)
	(cond ((null tokl) (princ "missing name") (nterpri) (return nil))
	      ((eq (car tokl) 'object) (return (oshow1 (cdr tokl))))
	      ((eq (car tokl) 'morphism) (return (oshow2)))
	      ((and (null (get (car tokl) 'morphism))
		    (get (car tokl) 'object))
	       (return (oshow1 tokl))))
	(setq $var 0)
	(setq ml (errset (morph0 tokl t) $erflag))
	(cond ((atom ml) (princ "Syntax error: ")
			 (oprinl $error)
			 (nterpri)
			 (oerrposprint tokl $errpos)
			 (return nil)))
	(setq dom (caddar ml))
	(setq codom (car (cdddar ml)))
	(setq ml (caar ml))
	(cond ((eq $system-mode 'window)
	       (w-select $w-show t)
	       (wlscroll $w-show 'bottom)
	       (whprinc $w-show "- Show Morphism -")
	       (nterpri)))
	(setq $ln (- (w-length $w-show) 10))
	(pmorphl ml -1)
	(nterpri)
	(princ "    : ")
	(pobj dom)
	(princ " -> ")
	(pobj codom)
	(nterpri)
	(return t))))

(def oshow1
 (lambda (tokl)
  (prog (obj objnm l iol ol)
	(cond ((eq $system-mode 'window)
	       (wpull $w-o-list)
	       (w-select $w-command nil))
	      ((null tokl)
	       (cond ((null (setq ol $object))
		      (princ "no objecs declared"))
		     (t (princ "objects: ")
			(oprinl (reverse (setq ol $object)))))
	       (nterpri)
	       (pp-while ol
			 (cond ((get (car ol) 'incorrect)
				(pp-cons (car ol) iol)))
			 (pp-cdr ol))
	       (cond (iol (princ "reparse need: ")
			  (oprinl iol)
			  (nterpri)))))
	(cond ((null tokl) (return t)))
	(setq objnm (car tokl))
	(setq obj (get objnm 'object))
	(cond ((null obj) (oprinl `("object" ,objnm "does not exist"))
			  (nterpri)
			  (return nil)))
	(cond ((eq $system-mode 'window)
	       (w-select $w-show t)
	       (wlscroll $w-show 'bottom)
	       (whprinc $w-show "- Show Object -")
	       (nterpri)))
	(setq $ln (- (w-length $w-show) 10))
	(princ (car obj))
	(princ " object ")
	(princ objnm)
	(setq l (cadr obj))
	(cond (l (princ (rlpar))
		 (princ (car l))
		 (pp-cdr l)
		 (pp-while l (princ (rcomma)) (princ (car l)) (pp-cdr l))
		 (princ (rrpar))))
	(princ " with ")
	(princ (caddr obj))
	(setq l (cadddr obj))
	(cond (l (princ " is")))
	(nterpri)
	(pp-while l
		  (princ "    ")
		  (pmobj (caar l) (cadar l) (caddar l))
		  (nterpri)
		  (pp-cdr l))
	(princ "end object")
	(nterpri))))

(def oshow2
 (lambda ()
  (prog (m x unit univ aux im)
	(cond ((null (setq m $morphism))
	       (princ "no morphisms declared")
	       (nterpri)
	       (return t))
	      ((eq $system-mode 'window)
	       (wpull $w-m-list)
	       (w-select $w-show t)
	       (wlscroll $w-show 'bottom)
	       (whprinc $w-show "- Morphisms -")
	       (nterpri)))
   aa	(cond ((null m) (go bb))
	      ((null (setq x (get (car m) 'morphism))) nil)
	      ((eq (setq x (car x)) 'unit) (pp-cons (car m) unit))
	      ((eq x 'factorizer) (pp-cons (car m) univ))
	      ((eq x 'auxiliary)
	       (pp-cons (car m) aux)
	       (cond ((get (car m) 'incorrect) (pp-cons (car m) im)))))
	(setq m (cdr m))
	(go aa)
   bb	(cond (unit (princ "units: ") (oprinl unit) (nterpri)))
	(cond (univ (princ "factorizers: ") (oprinl univ) (nterpri)))
	(cond (aux (princ "auxiliaries: ") (oprinl aux) (nterpri)))
	(cond (im (princ "reparse need: ") (oprinl im) (nterpri)))
	(return t))))

(def osave
 (lambda (tokl)
  (prog (l fn port)
	(cond ((null tokl) (princ "no file name") (nterpri) (return nil)))
	(setq fn (concatl tokl))
	(setq port (fileopen fn "w"))
	(print `(setq $object (quote ,$object)) port)
	(terpri port)
	(print `(setq $morphism (quote ,$morphism)) port)
	(terpri port)
	(setq l (reverse $object))
   aa   (cond ((null l) (go bb)))
	(print `(defprop ,(car l) ,(get (car l) 'object) object) port)
	(terpri port)
	(print `(defprop ,(car l) ,(get (car l) 'morphism) morphism) port)
	(terpri port)
	(setq l (cdr l))
	(go aa)
   bb	(setq l (reverse $morphism))
   cc	(cond ((null l) (go dd)))
	(print `(defprop ,(car l) ,(get (car l) 'morphism) morphism) port)
	(terpri port)
	(setq l (cdr l))
	(go cc)
   dd   (print `(setq $aliases (quote ,$aliases)) port)
	(terpri port)
	(print `(setq $keypad (quote ,$keypad)) port)
	(terpri port)
	(close port)
	(princ "saved")
	(nterpri)
	(return t))))

(def oload
 (lambda (tokl)
  (prog (fn x pt)
	(cond ((null tokl) (princ "no file name") (nterpri) (return nil)))
	(setq fn (concatl tokl))
	(cond ((null (probef fn)) (oprinl `("file" ,fn "does not exist"))
				  (nterpri)
				  (return nil)))
	(oclear)
	(setq pt (fileopen fn "r"))
   aa   (setq x (read pt))
	(cond ((null x) (close pt) (return t)))
	(eval x)
	(go aa))))

(def oclear
 (lambda ()
  (prog (l)
	(setq l $object)
	(setq $object nil)
   aa   (cond ((null l) (go bb)))
	(remprop (car l) 'object)
	(remprop (car l) 'morphism)
	(remprop (car l) 'incorrect)
	(setq l (cdr l))
	(go aa)
   bb   (setq l $morphism)
	(setq $morphism nil)
   cc	(cond ((null l) (return t)))
	(remprop (car l) 'morphism)
	(remprop (car l) 'incorrect)
	(setq l (cdr l))
	(go cc))))

; trace

(declare (special $trace $tplvl))
(setq $trace nil)
(setq $tplvl 2)

(def oset
 (lambda (tokl)
  (cond ((null tokl)
	 (cond ((eq $system-mode 'window)
		(w-select $w-show t)
		(wlscroll $w-show 'bottom)
		(whprinc $w-show "- Flags -")
		(nterpri)))
	 (princ "simplification mode: ")
	 (princ $simp-mode)
	 (nterpri)
	 (princ "trace ")
	 (cond ($trace (princ "on")) (t (princ "off")))
	 (nterpri)
	 (princ "trace print depth ")
	 (princ $tplvl)
	 (nterpri)
	 (cond ((eq $system-mode 'window) (princ "window mode"))
	       (t (princ "line input mode")))
	 (nterpri)
	 (cond ((eq $arrow-dir 'normal) (princ "print arrow normal"))
	       (t (princ "print arrow reversed")))
	 (nterpri))
	((or (eq (car tokl) 'simp) (eq (car tokl) 'eval))
	 (cond ((or (null (cdr tokl)) (eq (cadr tokl) 'lazy))
		(setq $simp-mode 'lazy))
	       (t (setq $simp-mode 'full))))
	((eq (car tokl) 'window)
	 (cond ((or (null (cdr tokl)) (eq (cadr tokl) 'on))
		(cond ((not (eq $system-mode 'window))
		       (wc-init)
		       (wc-update-list))))
	       ((eq $system-mode 'window) (wc-end))))
	((eq (car tokl) 'trace)
	 (setq $trace (or (null (cdr tokl)) (eq (cadr tokl) 'on))))
	((eq (car tokl) 'print)
	 (cond ((null (cdr tokl)) (setq $tplvl -1))
	       (t (setq $tplvl (ogtnum (cadr tokl))))))
	((eq (car tokl) 'arrow)
	 (cond ((or (null (cdr tokl)) (eq (cadr tokl) 'normal))
		(setq $arrow-dir 'normal))
	       (t (setq $arrow-dir 'reverse))))
	(t (princ "Illegal parameter for set.") (nterpri)))))

(def ogtnum
 (lambda (tok)
  (prog (l n)
	(setq l (exploden tok))
	(setq n 0)
   aa   (cond ((or (null l) (not (rdigit (car l))))
	       (cond ((zerop n) (return nil)) (t (return n)))))
	(setq n (plus (times n 10) (difference (car l) (rascii \0))))
	(setq l (cdr l))
	(go aa))))

; delete

(def odel
 (lambda (tokl)
  (prog (mn bd)
	(cond ((null tokl) (princ "No name") (nterpri) (return nil)))
	(setq mn (car tokl))
	(cond ((get mn 'object) (return (odelo mn t))))
	(setq bd (get mn 'morphism))
	(cond ((null bd)
	       (princ mn) (princ " undefined") (nterpri) (return nil))
	      ((not (eq (car bd) 'auxiliary))
	       (oprinl `(,(car bd)
			 ,mn
			 "cannot be deleted separately from"
			 ,(cadr bd)))
	       (nterpri)
	       (return nil)))
	(return (odelm mn t)))))

(def odelm
 (lambda (aux dflg)
  (prog (depl)
	(setq depl (odelmd aux nil))
	(cond (depl (cond ((not dflg) (princ "Modified ")))
		    (princ aux)
		    (princ " was used in: ")
		    (oprinl depl)
		    (nterpri)))
   aa   (cond (depl (putprop (car depl) t 'incorrect)
		    (setq depl (cdr depl))
		    (go aa)))
	(remprop aux 'morphism)
	(remprop aux 'incorrect)
	(setq $morphism (odell $morphism (list aux)))
	(cond (dflg (oprinl `(,aux "deleted")) (nterpri)))
	(return t))))

(def odelmd
 (lambda (aux depl)
  (prog (ml bd)
	(setq ml $morphism)
   aa   (cond ((null ml) (return depl))
	      ((memq (car ml) depl))
	      ((and (eq (car (setq bd (get (car ml) 'morphism))) 'auxiliary)
		    (odelmdml (caddr bd) aux))
	       (setq depl (odelmd (car ml) (cons (car ml) depl)))))
	(setq ml (cdr ml))
	(go aa))))

(def odelmdmll
 (lambda (mll aux)
  (cond ((null mll) nil)
	((odelmdml (car mll) aux) t)
	(t (odelmdmll (cdr mll) aux)))))

(def odelmdml
 (lambda (ml aux)
  (cond ((null ml) nil)
	((odelmdm (car ml) aux) t)
	(t (odelmdml (cdr ml) aux)))))

(def odelmdm
 (lambda (morph aux)
  (cond ((eq (car morph) 'factorizer) (odelmdmll (caddr morph) aux))
	((not (eq (car morph) 'auxiliary)) nil)
	(t (eq (cadr morph) aux)))))

(def odell
 (lambda (l a)
  (cond ((null l) nil)
	((memq (car l) a) (odell (cdr l) a))
	(t (cons (car l) (odell (cdr l) a))))))

(def odelo
 (lambda (obj dflg)
  (prog (objl depl bd unitl ml)
	(setq objl $object)
   aa   (cond ((null objl) (go a1))
	      ((and (not (eq (car objl) obj)) (odelod (car objl) obj))
	       (setq depl (cons (car objl) depl))))
	(setq objl (cdr objl))
	(go aa)
   a1   (setq ml $morphism)
   a2   (cond ((null ml) (go bb))
	      ((and (eq (car (setq bd (get (car ml) 'morphism)))
			'auxiliary)
		    (odelodml (caddr bd) obj))
	       (setq depl (odelmd (car ml) (cons (car ml) depl)))))
	(setq ml (cdr ml))
	(go a2)
   bb	(cond (depl (cond ((not dflg) (princ "Modified ")))
		    (princ obj)
		    (princ " was used in: ")
		    (oprinl (reverse depl))
		    (nterpri)))
   b0   (cond (depl (putprop (car depl) t 'incorrect)
		    (setq depl (cdr depl))
		    (go b0)))
	(setq bd (get obj 'object))
	(remprop obj 'object)
	(remprop obj 'incorrect)
	(setq $object (odell $object (list obj)))
	(remprop obj 'morphism)
	(setq ml (list (caddr bd)))
	(remprop (caddr bd) 'morphism)
	(setq unitl (cadddr bd))
    cc  (cond ((null unitl)
	       (cond (dflg (setq $morphism (odell $morphism ml))
			   (princ obj)
			   (princ " deleted")
			   (nterpri)))
	       (return t)))
	(setq ml (cons (caar unitl) ml))
	(remprop (caar unitl) 'morphism)
	(setq unitl (cdr unitl))
	(go cc))))

(def odelod
 (lambda (obj dobj)
  (prog (unitl)
	(setq unitl (cadddr (get obj 'object)))
   aa   (cond ((null unitl) (return nil))
	      ((or (odelodo (cadar unitl) dobj)
		   (odelodo (caddar unitl) dobj))
	       (return t)))
	(setq unitl (cdr unitl))
	(go aa))))

(def odelodo
 (lambda (dom dobj)
  (cond ((atom dom) nil)
	((eq (cadr dom) dobj) t)
	(t (odelodol (caddr dom) dobj)))))

(def odelodol
 (lambda (doml dobj)
  (and doml (or (odelodo (car doml) dobj) (odelodol (cdr doml) dobj)))))

(def odelodmll
 (lambda (mll dobj)
  (and mll (or (odelodml (car mll) dobj) (odelodmll (cdr mll) dobj)))))

(def odelodml
 (lambda (ml dobj)
  (and ml (or (odelodm (car ml) dobj) (odelodml (cdr ml) dobj)))))

(def odelodm
 (lambda (m dobj)
  (cond ((eq (car m) 'unit) (eq (cadr (get (cadr m) 'morphism)) dobj))
	((eq (car m) 'factorizer)
	 (or (eq (cadr (get (cadr m) 'morphism)) dobj)
	     (odelodmll (caddr m) dobj)))
	(t nil))))

(def oinco
 (lambda (o)
  (prog (unitl)
	(setq unitl (cadddr (get o 'object)))
   aa   (cond ((null unitl) (return nil))
	      ((or (oincdo (cadar unitl)) (oincdo (caddar unitl)))
	       (return t)))
	(setq unitl (cdr unitl))
	(go aa))))

(def oincdo
 (lambda (oo)
  (cond ((atom oo) nil)
	((get (cadr oo) 'incorrect) t)
	(t (oincdol (caddr oo))))))

(def oincdol
 (lambda (ool)
  (and ool (or (oincdo (car ool)) (oincdol (cdr ool))))))

(def oincml
 (lambda (ml)
  (and ml (or (oincm (car ml)) (oincml (cdr ml))))))

(def oincm
 (lambda (m)
  (cond ((eq (car m) 'unit)
	 (get (caddr (get (cadr m) 'morphism)) 'incorrect))
	((eq (car m) 'factorizer)
	 (or (get (caddr (get (cadr m) 'morphism)) 'incorrect)
	     (oincmll (caddr m))))
	((eq (car m) 'auxiliary) (get (cadr m) 'incorrect))
	(t nil))))

(def oincmll
 (lambda (mll)
  (and mll (or (oincml (car mll)) (oincmll (cdr mll))))))

(def oexpand
 (lambda (tokl)
  (prog (aux m ml)
   aa	(cond ((null tokl) (return nil)))
	(setq aux (car tokl))
	(setq tokl (cdr tokl))
	(cond ((null (setq m (get aux 'morphism)))
	       (oprinl `(,aux "undefined"))
	       (nterpri)
	       (go aa))
	      ((not (eq (car m) 'auxiliary))
	       (oprinl `(,aux "is not an auxiliary morphism"))
	       (nterpri)
	       (go aa)))
	(oexp aux (caddr m) (cadddr m))
	(oprinl `(,aux "expanded"))
	(nterpri)
	(go aa))))

(def oexp
 (lambda (aux m vl)
  (prog (ml dc)
	(setq ml $morphism)
	(go aa)
   bb   (setq ml (cdr ml))
   aa   (cond ((null ml) (return nil))
	      ((not (eq (car (setq dc (get (car ml) 'morphism))) 'auxiliary))
	       (go bb)))
	(putprop (car ml)
		 (list 'auxiliary
		       '*auxiliary
		       (oexpml (caddr dc) aux m vl)
		       (cadddr dc)
		       (car (cddddr dc))
		       (cadr (cddddr dc)))
		 'morphism)
	(go bb))))

(def oexpmll
 (lambda (mll aux m vl)
  (cond ((null mll) nil)
	(t (cons (oexpml (car mll) aux m vl) (oexpmll (cdr mll) aux m vl))))))

(def oexpml
 (lambda (ml aux m vl)
  (prog (rtv morph)
   aa   (cond ((null ml) (return (nreverse rtv)))
	      ((and (eq (car (setq morph (car ml))) 'auxiliary)
		    (eq (cadr morph) aux))
	       (setq rtv (revapp m rtv)))
	      ((eq (car morph) 'factorizer)
	       (setq rtv
		     (cons (list 'factorizer
				 (cadr morph)
				 (oexpmll (caddr morph) aux m vl))
			   rtv)))
	      (t (setq rtv (cons morph rtv))))
	(setq ml (cdr ml))
	(go aa))))

;; simplifier (evaluater)

(declare (special $simp-full $cnt))

(def simp
 (lambda (lazy tokl)
  (prog (m sm y edit win)
	(cond ((and tokl (memq (car tokl) '(lazy full)))
	       (setq $simp-full (eq (car tokl) 'full))
	       (pp-cdr tokl))
	      (t (setq $simp-full (eq $simp-mode 'full))))
	(cond ((not (setq edit (null tokl))) (go bb))
	      ((setq win (eq $system-mode 'window)) (we-new)))
   aa   (cond (win (w-select $w-edit t)
		   (cond ((= (wedit $w-edit) 3) (return nil)))
		   (setq $w-ll (we-get))
		   (rclear)
		   (setq tokl (rtokl 'window)))
	      (t (rclear)
		 (setq $rnl t)
		 (setq tokl (rtokl nil))
		 (wc-input nil "" "")))
	(rclear)
   bb	(setq m (errset (morph0 tokl nil) $erflag))
	(cond ((atom m) (princ "Syntax error: ")
			(oprinl $error)
			(nterpri)
			(oerrposprint tokl $errpos)
			(cond (edit (go aa)))
			(return nil)))
	(setq m (car m))
	(cond ((oincml (car m))
	       (princ "Something may have been changed. ")
	       (princ "Update must be done first.") (go zz)))
	(setq $cnt 0)
	(cond ($trace
	       (cond ((eq $system-mode 'window)
		      (w-select $w-show t)
		      (wlscroll $w-show 'bottom)
		      (whprinc $w-show "- Trace Output -")
		      (nterpri)))
	       (setq $ln (- (w-length $w-show) 10))))
	(setq sm (errset (simpl (reverse (car m)) nil 0) $erflag))
	(w-select $w-command nil)
	(cond ((atom sm) (princ "Error in evaluation.") (go zz)))
	(setq sm (car sm))
	(pmorphl sm $tplvl)
	(nterpri)
	(princ "    :")
	(pobj (caddr m))
	(princ " -> ")
	(pobj (cadddr m))
	(nterpri)
	(setq y (get 'it 'morphism))
	(cond ((or (null y) (eq (car y) 'auxiliary))
	       (oexp 'it (caddr y) (cadddr y))
	       (putprop 'it
			`(auxiliary *auxiliary ,sm nil ,(caddr m) ,(cadddr m))
			 'morphism)
	       (remprop 'it 'incorrect)
	       (cond ((not (memq 'it $morphism))
		      (setq $morphism (cons 'it $morphism))
		      (wc-update-list)))))
	(return t)
   zz	(nterpri)
	(return nil))))

(def revapp
 (lambda (a b)
  (cond ((null a) b) (t (revapp (cdr a) (cons (car a) b))))))

(def revcomp
 (lambda (a b)
  (cond ((null a) b) (t (revcomp (cdr a) (mcomp1 (car a) b))))))

(def simpl
 (lambda (ml sml level)
  (prog (m mm obj objnm x)
   aa   (cond ($trace (princ $cnt)
		      (cond ((not (zerop level))
			     (princ "[")
			     (princ level)
			     (princ "]")))
		      (princ ":")
		      (cond ((eq $arrow-dir 'reverse)
			     (cond (sml (pmorphl sml $tplvl)))
			     (princ "*")
			     (cond (ml (pmorphl (reverse ml) $tplvl))))
			    (t (cond (ml (pmorphl (reverse ml) $tplvl)))
			       (princ "*")
			       (cond (sml (pmorphl sml $tplvl)))))
		      (nterpri)
		      (setq $cnt (add1 $cnt))))
	(cond ((null ml)
	       (cond ((null sml) (return '((identity))))
		     (t (return sml)))))
	(setq m (car ml))
	(cond ((eq (car m) 'identity) (pp-cdr ml) (go aa))
	      ((eq (car m) 'auxiliary)
	       (setq mm (get (cadr m) 'morphism))
	       (setq ml (revcomp (caddr mm) (cdr ml)))
	       (go aa))
	      ((eq (car m) 'unit)
	       (setq objnm (cadr (get (cadr m) 'morphism)))
	       (setq obj (get objnm 'object))
	       (cond ((or (null sml) (eq (car obj) 'left))
		      (setq sml (mcomp1 m sml)) (pp-cdr ml) (go aa)))
	       (setq mm (suu 'right ml sml objnm obj level))
	       (setq sml (cadr mm))
	       (setq ml (car mm))
	       (go aa))
	      ((eq (car m) 'factorizer)
	       (setq objnm (cadr (get (cadr m) 'morphism)))
	       (setq obj (get objnm 'object))
	       (cond ((eq (car obj) 'right)
		      (cond ((null (cadddr obj)) (setq sml (list m)))
			    ((and $simp-full
				  (eq (car (cddddr obj)) 'productive))
			     (setq sml (sprod m sml objnm obj level)))
			    (t (setq sml (mcomp1 m sml))))
		      (pp-cdr ml)
		      (go aa)))
	       (setq mm (suu 'left ml sml objnm obj level))
	       (setq sml (cadr mm))
	       (setq ml (car mm))
	       (go aa))
	      (t (error 0))))))

(def sprod
 (lambda (m sml objnm obj level)
  (prog (para unitl argl mll mll2 ml)
	(setq para (cadr obj))
	(pp-while para
		  (pp-cons (list (car para) '((identity)) nil nil) argl)
		  (pp-cdr para))
	(pp-cons (list objnm sml nil nil) argl)
	(setq mll (caddr m))
	(setq unitl (cadddr obj))
   aa   (pp-while unitl
		  (setq ml (car (mfungt (cadar unitl) argl)))
		  (pp-cons (simpl (reverse (car mll)) ml (1+ level)) mll2)
		  (pp-cdr unitl)
		  (pp-cdr mll))
	(return (list (list 'factorizer (cadr m) (nreverse mll2)))))))

(def suu
 (lambda (rl ml sml objnm obj level)
  (prog (unit arrow F G right mm para argl Gml mll unitl)
	(cond ((null sml) (go zz))
	      ((setq right (eq rl 'right)) (setq unit (car ml)))
	      (t (setq unit (car sml))))
	(cond ((not (eq (car unit) 'unit)) (go zz)))
	(setq arrow (assq (cadr unit) (setq unitl (cadddr obj))))
	(cond (right (setq F (cadr arrow))
		     (setq G (caddr arrow))
		     (setq mm (suuget t F objnm sml level))
		     (cond ((null (car mm)) (setq sml (cadr mm)) (go zz)))
		     (setq sml (caddr mm)))
	      (t (setq F (caddr arrow))
		 (setq G (cadr arrow))
		 (setq mm (suuget nil F objnm ml level))
		 (cond ((null (car mm)) (setq ml (cadr mm)) (go zz)))
		 (setq ml (caddr mm))))
	(setq mm (cadr mm))
	(setq para (cadr obj))
	(pp-while para
		  (pp-cons (list (car para) '((identity)) nil nil) argl)
		  (pp-cdr para))
	(pp-cons (list objnm (list mm) nil nil) argl)
	(setq Gml (car (mfungt G argl)))
	(setq mll (caddr mm))
	(pp-while (not (eq (caar unitl) (cadr unit)))
		  (pp-cdr unitl)
		  (pp-cdr mll))
	(cond (right
	       (setq ml (revcomp Gml (cdr ml)))
	       (cond ((and $simp-full
			   (or (null sml) (equal sml '((identity))))
			   (eq (car (cddddr obj)) 'productive))
		      (setq sml (car mll)))
		     (t (setq ml (revcomp (car mll) ml)))))
	      (t (setq ml (revcomp Gml (revcomp (car mll) ml)))
		 (pp-cdr sml)))
	(return (list ml sml))
    zz  (return (list (cdr ml) (mcomp1 (car ml) sml))))))

(def suuget
 (lambda (right F objnm ml level)
  (prog (univ obj unitl mll mll2 para apara argl)
	(setq univ (car ml))
	(cond ((not (eq (car univ) 'factorizer)) (go zz))
	      ((eq F objnm)
	       (cond ((not (eq (cadr (get (cadr univ) 'morphism)) objnm))
		      (go zz)))
	       (return (list t (car ml) (mcomp1 '(identity) (cdr ml)))))
	      ((not (eq (cadr F) (cadr (get (cadr univ) 'morphism)))) (go zz)))
	(setq obj (get (cadr F) 'object))
	(cond ((null (cdr ml)) (go aa)))
	(setq unitl (cadddr obj))
	(setq para (cadr obj))
	(pp-while para
		  (pp-cons (list (car para) '((identity)) nil nil) argl)
		  (pp-cdr para))
	(pp-cons (list (cadr F) (cdr ml) nil nil) argl)
	(setq mll (caddr univ))
	(pp-while unitl
		  (cond (right
			 (setq ml (car (mfungt (cadar unitl) argl)))
			 (pp-cons (mcomp (car mll) ml) mll2))
			(t (setq ml (car (mfungt (caddar unitl) argl)))
			   (pp-cons (mcomp ml (car mll)) mll2)))
		  (pp-cdr unitl)
		  (pp-cdr mll))
	(setq univ (list 'factorizer (cadr univ) (nreverse mll2)))
   aa   (setq para (cadr obj))
	(setq apara (caddr F))
	(pp-while (zerop (ooccur (car apara) objnm))
		  (pp-cdr para)
		  (pp-cdr apara))
	(setq para (car para))
	(setq unitl (cadddr obj))
	(setq mll (caddr univ))
	(setq mll2 nil)
	(pp-while (not (eq (cond (right (caddar unitl)) (t (cadar unitl)))
			   para))
		  (pp-cdr unitl)
		  (pp-cons (car mll) mll2)
		  (pp-cdr mll))
	(setq ml (car mll))
	(pp-cdr mll)
	(cond ((cdr ml)
	       (cond (right
		      (cond ((not $simp-full)
			     (setq ml (simpl (reverse ml) nil (1+ level))))))
		     (t (setq ml (simpl ml nil (1+ level)))))))
	(setq ml (suuget right (car apara) objnm ml (1+ level)))
	(cond ((null (car ml))
	       (pp-cons (cadr ml) mll)
	       (pp-while mll2 (pp-cons (car mll2) mll) (pp-cdr mll2))
	       (return (list nil (list (list 'factorizer (cadr univ) mll))))))
	(pp-cons (caddr ml) mll)
	(pp-while mll2 (pp-cons (car mll2) mll) (pp-cdr mll2))
	(return (list t (cadr ml) (list (list 'factorizer (cadr univ) mll))))
   zz   (return (list nil ml)))))
	

;; windows

(def wcat
 (lambda ()
  (prog (tokl stand-alone atokl)
	(cond ((setq stand-alone (eq user-top-level 'wc))
	       (setq user-top-level nil)))
	(rclear)
	(cond ((eq $system-mode 'window) (wc-init)))
	(wc-update-list)
	(w-select $w-command nil)
	(princ "Categorical Programming Language (version 3)")
	(nterpri)
   aa   (setq $w-ll (wc-input $w-command "cpl>" "   >"))
	(setq $rtyi nil)
	(setq $reof nil)
	(setq $ln (- (w-length $w-command) 10))
	(cond ((and $w-ll (eq (car $w-ll) 'token)) (setq tokl (cdr $w-ll)))
	      (t (setq tokl (rtokl 'window))))
   bb	(cond ((null tokl) (go aa))
	      ((eq (car tokl) 'quit)
	       (cond ((eq $system-mode 'window) (wc-end)))
	       (princ "bye")
	       (nterpri)
	       (cond ((or (not stand-alone)
			  (and (cdr tokl) (eq (cadr tokl) 'lisp)))
		      (return nil))
		     (t (exit))))
	      ((eq (car tokl) 'edit) (wc-edit (cdr tokl)))
	      ((eq (car tokl) 'let) (maux (cdr tokl)))
	      ((eq (car tokl) 'delete)
	       (odel (cdr tokl)) (wc-update-list))
	      ((eq (car tokl) 'show) (oshow (cdr tokl)))
	      ((memq (car tokl) '(simp eval)) (simp (car tokl) (cdr tokl)))
	      ((eq (car tokl) 'save) (osave (cdr tokl)))
	      ((eq (car tokl) 'load)
	       (oload (cdr tokl)) (wc-update-list))
	      ((eq (car tokl) 'set) (oset (cdr tokl)))
	      ((eq (car tokl) 'read)
	       (oread (cdr tokl)) (wc-update-list))
	      ((eq (car tokl) 'new)
	       (oclear)
	       (princ "All the objects and morphisms are cleared")
	       (nterpri)
	       (wc-update-list))
	      ((eq (car tokl) 'expand) (oexpand (cdr tokl)))
	      ((eq (car tokl) 'help) (wc-help (cdr tokl)))
	      ((eq (car tokl) 'alias) (wc-alias (cdr tokl)))
	      ((eq (car tokl) 'keypad) (wc-keypad (cdr tokl)))
	      ((eq (car tokl) 'scroll) (wc-scroll (cdr tokl)))
	      ((eq (car tokl) 'diagram) (wc-diagram (cdr tokl)))
	      ((setq atokl (assq (car tokl) $aliases))
	       (setq tokl (append (cdr atokl) (cdr tokl)))
	       (go bb))
	      (t (princ "no command: ")
	         (princ (car tokl))
		 (princ "   type `help' for help")
		 (nterpri)))
	(go aa))))

(def wc-scroll
 (lambda (tokl)
  (prog (n)
	(w-select $w-show t)
	(cond ((null tokl) (wlscroll $w-show 'bottom))
	      ((eq (car tokl) '|-|)
	       (cond ((null (cdr tokl)) (wlscroll $w-show 'top))
		     (t (wlscroll $w-show (minus (ogtnum (cadr tokl)))))))
	      (t (wlscroll $w-show (ogtnum (car tokl)))))
	(return t))))

(def wc-alias
 (lambda (tokl)
  (prog (aliases naliases)
	(cond ((null tokl)
	       (cond ((eq $system-mode 'window)
		      (w-select $w-show t)
		      (wlscroll $w-show 'bottom)
		      (whprinc $w-show "- Aliases -")
		      (nterpri)))
	       (setq aliases $aliases)
	       (pp-while aliases
			 (oprinl (car aliases))
			 (nterpri)
			 (pp-cdr aliases))
	       (return t)))
	(setq aliases $aliases)
	(pp-while (and aliases (not (eq (caar aliases) (car tokl))))
		  (pp-cons (car aliases) naliases)
		  (pp-cdr aliases))
	(cond (aliases (setq $aliases (append naliases (cdr aliases)))))
	(cond ((null (cdr tokl)) (return t)))
	(setq $aliases (cons tokl $aliases))
	(return t))))

(def wc-keypad
 (lambda (tokl)
  (prog (keypad nkeypad)
	(cond ((null tokl)
	       (cond ((eq $system-mode 'window)
		      (w-select $w-show t)
		      (wlscroll $w-show 'bottom)
		      (whprinc $w-show  "- Keypad -")
		      (nterpri)))
	       (setq keypad $keypad)
	       (pp-while keypad
			 (princ (caar keypad))
			 (princ ": ")
			 (oprinl (cdar keypad))
			 (nterpri)
			 (pp-cdr keypad))
	       (return t)))
	(setq keypad $keypad)
	(pp-while (and keypad (not (eq (caar keypad) (car tokl))))
		  (pp-cons (car keypad) nkeypad)
		  (pp-cdr keypad))
	(cond (keypad (setq $keypad (append nkeypad (cdr keypad)))))
	(cond ((null (cdr tokl)) (return t)))
	(setq $keypad (cons tokl $keypad))
	(return t))))

(def wc-init
 (lambda ()
  (prog ()
	(setq $system-mode 'window)
	(wraw)
	(setq $w-edit (wmake 5 5 10 40 0))
	(wheading $w-edit "EDIT")
	(setq $w-diagram (wmake 1 1 16 63 20))
	(wheading $w-diagram "DIAGRAM EDITOR")
	(setq $w-show (wmake 1 1 16 63 20))
	(wheading $w-show "SHOW")
	(wscroll $w-show t)
	(setq $w-m-list (wmake 66 1 16 12 5))
	(wheading $w-m-list "MORPHISM")
	(wscroll $w-m-list t)
	(setq $w-o-list (wmake 66 1 16 12 5))
	(wheading $w-o-list "OBJECT")
	(wscroll $w-o-list t)
	(setq $w-command (wmake 1 19 4 77 10))
	(wheading $w-command "COMMAND")
	(wscroll $w-command t)
	(return nil))))

(def wc-end
 (lambda ()
  (prog ()
	(wcooked)
	(wdelete $w-edit)
	(setq $w-edit nil)
	(wdelete $w-diagram)
	(setq $w-diagram nil)
	(wdelete $w-show)
	(setq $w-show nil)
	(wdelete $w-m-list)
	(setq $w-m-list nil)
	(wdelete $w-o-list)
	(setq $w-o-list nil)
	(wdelete $w-command)
	(setq $w-command nil)
	(cond ($w-help-on (wdelete $w-help)))
	(setq $w-help nil)
	(setq $w-help-on nil)
	(setq $system-mode 'line)
	(return nil))))

(def w-select
 (lambda (w pull)
	 (cond ((eq $system-mode 'window)
		(cond (pull (wpull w)) (t (wselect w)))))))

(def w-length
 (lambda (w) (cond ((eq $system-mode 'window) (wlength w)) (t 80))))

(def wc-input
 (lambda (w prmpt prmpt2)
  (prog (ll l ch echo fn)
	(setq echo (eq $system-mode 'window))
	(w-select w nil)
   bb	(cond (prmpt (princ prmpt)))
	(setq prmpt prmpt2)
   aa   (setq ch (tyi))
	(cond ((and (<= 32 ch) (< ch 127))
	       (cond (echo (tyo ch)))
	       (setq l (cons ch l)))
	      ((and echo (or (= ch 8) (= ch 127)))
	       (cond (l (setq l (cdr l))
			(tyo 8)
			(tyo 32)
			(tyo 8))))
	      ((and echo (= ch 27))
	       (setq fn (assq (ascii (tyi)) $keypad))
	       (cond ((null fn) (go aa)))
	       (pp-while l (tyo 8) (tyo 32) (tyo 8) (pp-cdr l))
	       (oprinl (cdr fn))
	       (nterpri)
	       (return (cons 'token (cdr fn))))
	      ((or (= ch 13) (= ch 10))
	       (cond (echo (nterpri)))
	       (cond ((and l (= (car l) (rascii |&|)))
		      (setq ll (cons (nreverse (cdr l)) ll))
		      (setq l nil)
		      (go bb)))
	       (return (nreverse (cons (nreverse l) ll)))))
	(go aa))))

(def wc-edit
 (lambda (tokl)
  (prog (x ok win)
	(cond ((not (setq win (eq $system-mode 'window)))
	       (cond (tokl (princ
			    "You can't modify things unless in window mode")
			   (nterpri)
			   (return nil)))
	       (go aa))
	      ((null tokl) nil)
	      ((eq (car tokl) 'new) (we-new))
	      ((get (car tokl) 'object)
	       (we-put (wc-object (car tokl))))
	      ((setq x (get (car tokl) 'morphism))
	       (cond ((eq (car x) 'auxiliary)
		      (we-put (wc-morph (car tokl) x)))
		     (t (princ "Auxiliary morphism name please")
			(nterpri)
			(return nil))))
	      (t (princ "Object or morphism name please")
		 (nterpri)
		 (return nil)))
   aa   (cond (win (w-select $w-edit t)
		   (cond ((= (wedit $w-edit) 3) (return nil)))
		   (setq $w-ll (we-get))
		   (rclear)
		   (setq tokl (rtokl 'window)))
	      (t (rclear)
		 (setq $rnl t)
		 (setq tokl (rtokl nil))
		 (wc-input nil "" "")))
	(rclear)
	(w-select $w-command nil)
	(cond ((null tokl) (princ "Null input") (nterpri) (return nil))
	      ((or (eq (car tokl) 'right) (eq (car tokl) 'left))
	       (setq ok (object tokl)))
	      ((eq (car tokl) 'let) (setq ok (maux (cdr tokl))))
	      (t (princ "Object or morphism declaration please")
		 (nterpri)
		 (oerrposprint tokl tokl)
		 (go aa)))
	(cond ((null ok) (go aa)))
	(wc-update-list)
	(return t))))

(def wc-update-list
 (lambda ()
  (prog (ol ml)
	(cond ((not (eq $system-mode 'window)) (return nil)))
	(wclall $w-o-list)
	(setq ol $object)
   aa   (cond (ol (cond ((get (car ol) 'incorrect) (princ "*")))
		  (princ (car ol))
		  (nterpri)
		  (setq ol (cdr ol))
		  (go aa)))
	(wclall $w-m-list)
	(setq ml $morphism)
   bb   (cond ((null ml) (return nil))
	      ((eq (car (get (car ml) 'morphism)) 'auxiliary)
	       (cond ((get (car ml) 'incorrect) (princ "*")))
	       (princ (car ml))
	       (nterpri)))
	(setq ml (cdr ml))
	(go bb))))

(declare (special $wc-l $wc-ll $wc-c))

(def wc-object
 (lambda (objnm)
  (prog (obj l)
	(setq $wc-l nil)
	(setq $wc-c 0)
	(setq $wc-ll nil)
	(setq obj (get objnm 'object))
	(wc-princ (car obj))
	(wc-princ " object ")
	(wc-princ objnm)
	(setq l (cadr obj))
	(cond ((null l) (go bb)))
	(wc-princ (rlpar))
   aa   (wc-princ (car l))
	(setq l (cdr l))
	(cond (l (wc-princ (rcomma)) (go aa)))
	(wc-princ (rrpar))
   bb   (wc-princ " with ")
	(wc-princ (caddr obj))
	(setq l (cadddr obj))
	(cond (l (wc-princ " is")))
	(wc-terpri)
   cc   (cond ((null l) (wc-princ "end object")
			(wc-terpri)
			(return (nreverse $wc-ll))))
	(wc-princ "    ")
	(wc-princ (caar l))
	(wc-princ " : ")
	(wc-pobj (cadar l))
	(wc-princ " -> ")
	(wc-pobj (caddar l))
	(wc-terpri)
	(setq l (cdr l))
	(go cc))))

(def wc-pobj
 (lambda (o)
  (prog ()
	(cond ((atom o) (wc-princ o) (return nil)))
	(wc-princ (cadr o))
	(setq o (caddr o))
	(cond ((null o) (return nil)))
	(wc-princ (rlpar))
   aa   (wc-pobj (car o))
	(setq o (cdr o))
	(cond (o (wc-princ (rcomma)) (go aa)))
	(wc-princ (rrpar))
	(return nil))))

(def wc-morph
 (lambda (mn dc)
  (prog ()
	(setq $ln (- (w-length $w-edit) 10))
	(setq $wc-ll nil)
	(setq $wc-l nil)
	(setq $wc-c 0)
	(wc-princ "let ")
	(wc-princ mn)
	(wc-princ " = ")
	(wc-pmorphl (caddr dc))
	(wc-terpri)
	(return (nreverse $wc-ll)))))

(def wc-pmorph
 (lambda (m)
  (progn
   (cond ((> $wc-c $ln) (wc-terpri) (wc-princ "    ")))
   (cond ((eq (car m) 'identity) (wc-princ "id"))
	 ((or (eq (car m) 'unit) (eq (car m) 'auxiliary))
	  (wc-princ (cadr m)))
	 ((eq (car m) 'factorizer)
	  (prog (mll)
	        (wc-princ (cadr m))
	        (setq mll (caddr m))
	        (cond ((null mll) (return nil)))
	        (wc-princ (rlpar))
	   aa   (wc-pmorphl (car mll))
	        (setq mll (cdr mll))
	        (cond ((null mll) (wc-princ (rrpar)) (return nil)))
	        (wc-princ (rcomma))
	        (go aa)))))))

(def wc-pmorphl
 (lambda (ml)
  (cond ((null ml))
	((eq $arrow-dir 'reversed)
	 (cond ((cdr ml) (wc-pmorphl (cdr ml)) (wc-princ (rrcomp))))
	 (wc-pmorph (car ml)))
	(t (wc-pmorph (car ml))
	   (cond ((cdr ml) (wc-princ (rcomp)) (wc-pmorphl (cdr ml))))))))

(def wc-princ
 (lambda (x)
  (prog (l)
	(setq l (exploden x))
	(setq $wc-l (revapp (exploden x) $wc-l))
	(setq $wc-c (+ $wc-c (length l)))
	(return l))))

(def wc-terpri
 (lambda ()
  (progn (setq $wc-ll (cons (nreverse $wc-l) $wc-ll))
	 (setq $wc-c 0)
	 (setq $wc-l nil))))

;; help

(declare (special $help-message))

(def wc-help
 (lambda (tokl)
  (prog (hm subject)
	(cond ((null tokl) (setq subject 'general))
	      ((eq (car tokl) 'off)
	       (cond ((and (eq $system-mode 'window) $w-help-on)
		      (wdelete $w-help)
		      (setq $w-help-on nil)))
	       (return nil))
	      (t (setq subject (car tokl))))
	(setq hm $help-message)
   aa   (cond ((null hm)
	       (oprinl `("no help for" subject))
	       (nterpri)
	       (return nil))
	      ((not (eq (caar hm) subject))
	       (setq hm (cdr hm))
	       (go aa)))
	(setq hm (cadar hm))
	(cond ((eq $system-mode 'window) (wc-help-window hm)))
   bb   (cond ((null hm) (return nil)))
	(princ (car hm))
	(setq hm (cdr hm))
	(cond ((or hm (not (eq $system-mode 'window))) (nterpri)))
	(go bb))))

(def wc-help-window
 (lambda (msgl)
  (prog (col row l n)
	(cond ($w-help-on (wdelete $w-help)))
	(setq col 4)
	(setq row 0)
	(setq l msgl)
   aa   (cond ((null l) (go bb))
	      ((> (setq n (flatc (car l))) col) (setq col n)))
	(setq l (cdr l))
	(setq row (1+ row))
	(go aa)
   bb   (setq $w-help (wmake (- 70 col) 3 row (1+ col) 0))
	(wheading $w-help "HELP")
	(w-select $w-help nil)
	(setq $w-help-on t)
	(return nil))))

(def wc-diagram
 (lambda (tokl)
  (prog (name sl ml y)
	(cond ((null tokl) (setq name nil))
	      ((not (eq $system-mode 'window))
	       (cond ((get (car tokl) 'incorrect)
		      (princ "Difinition is incorrect")
		      (nterpri)
		      (return nil)))
	       (setq sl (de-print (car tokl)))
	       (cond ((null sl)
		      (princ "An auxiliary morphism name should be given")
		      (nterpri)
		      (return nil)))
	       (de-display sl 0 0 0 32767 nil)
	       (return t))
	      (t (setq name (car tokl))
		 (cond ((get name 'incorrect)
			(princ "Difinition is incorrect")
			(nterpri)
			(return nil)))))
	(setq ml (dedit $w-diagram name))
	(wselect $w-command)
	(setq name (car ml))
	(setq ml (cdr ml))
	(cond ((null ml) (return nil))
	      ((and (setq y (get name 'morphism))
		    (not (eq (car y) 'auxiliary)))
	       (oprinl `("morphism" ,name "is already defined as" ,(car y)))
	       (nterpri)
	       (return nil))
	      (y (odelm name nil)))
	(putprop name
		 (cons 'auxiliary
		       (cons '*auxiliary ml))
		 'morphism)
	(cond ((oincml (car ml)) (putprop name t 'incorrect))
	      (t (remprop name 'incorrect)))
	(cond ((not (memq name $morphism))
	       (setq $morphism (cons name $morphism))))
	(pmobj name (caddr ml) (cadddr ml))
	(princ " defined")
	(nterpri))))

;; bootsrap
(load "wdia")
(load "wcathelp")
(load "wmlib")

(def wc
 (lambda ()
  (prog (args debug)
	(tyi)
	(cond ((eq user-top-level 'wc)
	       (setq args (command-line-args))
	       (pp-while args
			 (cond ((memq (car args) '(line window))
				(setq $system-mode (car args)))
			       ((eq (car args) 'debug)
				(setq $erflag t)
				(setq debug t)))
			 (pp-cdr args))
	       (cond ((not debug) (sstatus translink 'on)))))
	(wcat))))

(cond ($stand-alone (setq user-top-level 'wc)))
